from asyncio import coroutine
from typing import Dict, Text, Any
from unittest.mock import Mock, MagicMock

from _pytest.monkeypatch import MonkeyPatch

from pytest_sanic.utils import TestClient

from rasax.community.services.model_service import ModelService
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.stack_service import StackService


# noinspection PyPep8Naming,SpellCheckingInspection
def AsyncMagicMock(*args: Any, **kwargs: Any) -> MagicMock:
    """Awaitable version of `MagicMock`.

    It can be used exactly like `MagicMock` but replaces `await` expressions.
    """

    _coro = MagicMock(*args, **kwargs)
    kwargs.update(side_effect=coroutine(_coro))
    coroutine_function = MagicMock(*args, **kwargs)
    coroutine_function.coro = _coro
    return coroutine_function


async def test_model_training(
    test_client: TestClient, auth_header: Dict[Text, Text], monkeypatch: MonkeyPatch
):
    project = "default"

    # there aren't any connected `StackServices`, so let's mock them
    # noinspection PyTypeChecker
    stack_services = Mock(
        return_value={"worker": StackService(None, None, None, None, None)}
    )
    monkeypatch.setattr(
        SettingsService, "stack_services", stack_services,
    )

    # we're not actually training, so mock this process
    start_training_process = AsyncMagicMock(name="123")
    monkeypatch.setattr(
        StackService, "start_training_process", start_training_process,
    )

    # we need to artificially trigger a `FileExistsError`
    # this is what happens when a model exists
    model_path = "/some/random/path"
    save_trained_model = Mock(side_effect=FileExistsError(model_path))
    monkeypatch.setattr(
        ModelService, "save_trained_model", save_trained_model,
    )

    response = await test_client.post(
        f"/api/projects/{project}/models/jobs", headers=auth_header,
    )

    # now test for the expected status code
    assert response.status == 201
    assert await response.json() == {
        "info": "Model already exists.",
        "path": model_path,
    }
