from unittest.mock import Mock
from typing import Dict, Text, Any

import pytest
from pytest_sanic.plugin import TestClient
from _pytest.monkeypatch import MonkeyPatch
from sqlalchemy.orm import Session

from rasax.community.services.config_service import ConfigService, ConfigKey
from rasax.community import telemetry, config

TELEMETRY_URL = "/api/telemetry"


async def test_send_event(
    test_client: TestClient, auth_header: Dict[Text, Text], monkeypatch: MonkeyPatch,
):
    track = Mock()
    monkeypatch.setattr(telemetry, "track", track)

    data = {"event_name": "Test Event", "properties": {"foo": "bar"}}

    response = await test_client.post(TELEMETRY_URL, json=data, headers=auth_header)

    assert response.status == 204
    track.assert_called_once_with("Test Event", {"foo": "bar"}, None)


@pytest.mark.parametrize(
    "body",
    [
        {},
        {"event_name": "foo", "properties": {}},
        {"event_name": "foo", "properties": {}, "context": {}},
    ],
)
async def test_invalid_request(
    test_client: TestClient, auth_header: Dict[Text, Text], body: Dict[Text, Any],
):
    response = await test_client.post(TELEMETRY_URL, json=body, headers=auth_header)
    assert response.status == 400


async def test_disable_telemetry_already_disabled(
    monkeypatch: MonkeyPatch, test_client: TestClient, auth_header: Dict[Text, Text],
):
    mock = Mock(return_value=False)
    monkeypatch.setattr(telemetry, "is_telemetry_enabled", mock)

    response = await test_client.delete(TELEMETRY_URL, headers=auth_header)
    assert response.status == 200


async def test_disable_telemetry_local_mode(
    monkeypatch: MonkeyPatch, test_client: TestClient, auth_header: Dict[Text, Text],
):
    mock = Mock(return_value=True)
    monkeypatch.setattr(telemetry, "is_telemetry_enabled", mock)
    monkeypatch.setattr(config, "LOCAL_MODE", True)

    response = await test_client.delete(TELEMETRY_URL, headers=auth_header)
    assert response.status == 400


async def test_disable_telemetry(
    session: Session,
    monkeypatch: MonkeyPatch,
    test_client: TestClient,
    auth_header: Dict[Text, Text],
):
    mock = Mock(return_value=True)
    monkeypatch.setattr(telemetry, "is_telemetry_enabled", mock)
    monkeypatch.setattr(config, "LOCAL_MODE", False)

    response = await test_client.delete(TELEMETRY_URL, headers=auth_header)
    assert response.status == 200

    config_service = ConfigService(session)
    assert not config_service.get_value(ConfigKey.TELEMETRY_ENABLED)


async def test_read_config(
    monkeypatch: MonkeyPatch, test_client: TestClient, auth_header: Dict[Text, Text],
):
    mock = Mock(return_value=True)
    monkeypatch.setattr(telemetry, "is_telemetry_enabled", mock)

    response = await test_client.get(TELEMETRY_URL, headers=auth_header)
    data = await response.json()
    assert data["telemetry_enabled"]

    mock.return_value = False
    response = await test_client.get(TELEMETRY_URL, headers=auth_header)
    data = await response.json()
    assert not data["telemetry_enabled"]
